import logging
from typing import List, Optional

from .enums import Language, Platform, Title
from .object import Object

log: logging.Logger = logging.getLogger(__name__)


class Season(Object):
    """
    Represents a Call of Duty season object.

    Parameters
    ----------
    title : callofduty.Title
        Call of Duty title which the loot season originates.
    season : int
        Loot season number relative to title.
    platform : callofduty.Platform, optional
        Platform which the loot season is available on (default is PlayStation.)
    name : str, optional
        Official title of Season (default is None.)
    tiers : list, optional
        Array of Loot Items containing tier loot for the Season (default is an empty list.)
    chase : list, optional
        Array of Loot Items containing chase loot for the Season (default is an empty list.)
    language : callofduty.Language, optional
        Language which the loot data should be in (default is English.)
    """

    _type: str = "Season"

    def __init__(self, client, data: dict):
        super().__init__(client)

        self.title: Title = Title(data.pop("title"))
        self.season: int = data.pop("season")
        self.platform: Platform = Platform(data.pop("platform"))
        self.name: Optional[str] = data.pop("categoryTitle", None)
        self.tiers: List[LootItem] = []
        self.chase: List[LootItem] = []
        self.language: Language = Language(data.pop("language"))

        for tier in (_tiers := data.pop("tiers", [])) :
            self.tiers.append(LootItem(self, _tiers[tier]))

        for chase in (_chase := data.pop("chase", [])) :
            self.chase.append(LootItem(self, _chase[chase]))


class LootItem(Object):
    """
    Represents a Call of Duty loot item object.

    Parameters
    ----------
    id : str
        Internal name of the loot item.
    name : str
        Name of the loot item.
    category : str
        Type of the loot item.
    rarity : str
        Rarity of the loot item.
    tier : int
        Tier which the loot item is rewarded.
    image : str
        Image URL for the loot item.
    free : bool, optional
        Boolean indicating whether or not the loot item is available for free (default is False.)
    """

    _type: str = "LootItem"

    def __init__(self, client, data: dict):
        super().__init__(client)

        self.id: str = data.pop("name")
        self.name: str = data.pop("label")
        self.category: str = data.pop("type")
        self.rarity: str = data.pop("rarity")
        self.tier: int = int(data.pop("tier"))
        self.image: str = data.pop("image")
        self.free: bool = data.pop("free", False)



import logging
from typing import List, Union

from .enums import GameType, Language, Mode, Platform, TimeFrame, Title
from .leaderboard import Leaderboard
from .loot import Season
from .match import Match
from .player import Player
from .squad import Squad
from .utils import (
    StripHTML,
    VerifyGameType,
    VerifyLanguage,
    VerifyMode,
    VerifyPlatform,
    VerifyTimeFrame,
    VerifyTitle,
)

log = logging.getLogger(__name__)


class Client:
    """Client which manages communication with the Call of Duty API."""

    def __init__(self, http):
        self.http = http

    async def GetLocalize(self, language: Language = Language.English) -> dict:
        """
        Get the localized strings used by the Call of Duty Companion App
        and website.

        Parameters
        ----------
        language : callofduty.Language, optional
            Language to use for localization data (default is English.)

        Returns
        -------
        dict
            JSON data containing localized strings.
        """

        VerifyLanguage(language)

        web: dict = await self.http.GetWebLocalize(language.value)
        app: dict = await self.http.GetAppLocalize(language.value)

        return {**web, **app}

    async def GetNewsFeed(self, language: Language = Language.English) -> dict:
        """
        Get the Call of Duty franchise feed, includes blog posts and
        the Companion App message of the day.

        Parameters
        ----------
        language : callofduty.Language, optional
            Language to use for localization data (default is English.)

        Returns
        -------
        dict
            JSON data containing the Call of Duty franchise feed.
        """

        return await self.http.GetNewsFeed(language.value)

    async def GetFriendFeed(self, **kwargs) -> dict:
        """
        Get the Friend Feed of the authenticated Call of Duty player.

        Parameters
        ----------
        stripHTML : bool, optional
            Whether to strip the HTML formatting from rendered event strings (default is True.)

        Returns
        -------
        dict
            JSON data of the authenticated player's Friend Feed.
        """

        stripHTML: bool = kwargs.get("stripHTML", True)

        data: dict = (await self.http.GetFriendFeed())["data"]

        players: List[Player] = []
        for i in data["identities"]:
            _player: Player = Player(
                self, {"platform": i["platform"], "username": i["username"]}
            )
            players.append(_player)

        if stripHTML is True:
            for i in data["events"]:
                i["rendered"] = StripHTML(i["rendered"])

        return {
            "events": data["events"],
            "players": players,
            "metadata": data["metadata"],
        }

    async def GetMyIdentities(self) -> list:
        """
        Get the Title Identities for the authenticated Call of Duty player.

        Returns
        -------
        list
            Array of identities containing title, platform, username, and more.
        """

        data: dict = (await self.http.GetMyIdentities())["data"]

        identities: list = []

        for identity in data["titleIdentities"]:
            identities.append(
                {
                    "title": Title(identity.pop("title")),
                    "platform": Platform(identity.pop("platform")),
                    "username": identity["username"],
                    "activeDate": identity["activeDate"],
                    "activityType": identity["activityType"],
                }
            )

        return identities

    async def GetMyAccounts(self) -> List[Player]:
        """
        Get the linked Accounts for the authenticated Call of Duty player.

        Returns
        -------
        list
            Array of Player objects for the linked accounts.
        """

        data: dict = (await self.http.GetMyAccounts())["data"]

        accounts: List[Player] = []

        for account in data.keys():
            accounts.append(
                Player(
                    self, {"platform": account, "username": data[account]["username"]}
                )
            )

        return accounts

    async def GetMyFriends(self) -> List[Player]:
        """
        Get the Friends of the authenticated Call of Duty player.

        Returns
        -------
        list
            Array of Player objects for the friends.
        """

        data: dict = (await self.http.GetMyFriends())["data"]

        friends: List[Player] = []

        for friend in data["uno"]:
            friends.append(
                Player(
                    self,
                    {
                        "platform": friend["platform"],
                        "username": friend["username"],
                        "accountId": friend.get("accountId"),
                        "online": friend["status"]["online"],
                    },
                )
            )

        for _platform in data["firstParty"]:
            for friend in data["firstParty"][_platform]:
                i: list = friend.get("identities", [])
                identities: List[Player] = []

                for _platform in i:
                    identities.append(
                        Player(
                            self,
                            {
                                "platform": friend["identities"][_platform]["platform"],
                                "username": friend["identities"][_platform].get(
                                    "username"
                                ),
                                "accountId": friend["identities"][_platform][
                                    "accountId"
                                ],
                                "avatarUrl": friend["identities"][_platform].get(
                                    "avatarUrlLargeSsl"
                                ),
                            },
                        )
                    )

                friends.append(
                    Player(
                        self,
                        {
                            "platform": friend["platform"],
                            "username": friend["username"],
                            "accountId": friend.get("accountId"),
                            "avatarUrl": friend.get("avatarUrlLargeSsl"),
                            "online": friend["status"]["online"],
                            "identities": identities,
                        },
                    )
                )

        return friends

    async def GetMyFriendRequests(self) -> dict:
        """
        Get the incoming and outgoing Friend Requests for the authenticated
        Call of Duty player.

        Returns
        -------
        dict
            JSON data of the player's friend requests.
        """

        data: dict = (await self.http.GetMyFriends())["data"]

        incoming: List[Player] = []
        outgoing: List[Player] = []

        for request in data["incomingInvitations"]:
            incoming.append(
                Player(
                    self,
                    {
                        "platform": request["platform"],
                        "username": request["username"],
                        "accountId": request.get("accountId"),
                        "online": request["status"]["online"],
                    },
                )
            )

        for request in data["outgoingInvitations"]:
            outgoing.append(
                Player(
                    self,
                    {
                        "platform": request["platform"],
                        "username": request["username"],
                        "accountId": request.get("accountId"),
                        "online": request["status"]["online"],
                    },
                )
            )

        return {"incoming": incoming, "outgoing": outgoing}

    async def GetPlayer(self, platform: Platform, username: str) -> Player:
        """
        Get a Call of Duty player using their platform and username.

        Parameters
        ----------
        platform : callofduty.Platform
            Platform to get the player from.
        username : str
            Player's username for the designated platform.

        Returns
        -------
        object
            Player object for the requested player.
        """

        VerifyPlatform(platform)

        return Player(self, {"platform": platform.value, "username": username})

    async def SearchPlayers(
        self, platform: Platform, username: str, **kwargs
    ) -> List[Player]:
        """
        Search Call of Duty players by platform and username.

        Parameters
        ----------
        platform : callofduty.Platform
            Platform to get the players from.
        username : str
            Player's username for the designated platform.
        limit : int, optional
            Number of search results to return (default is None.)

        Returns
        -------
        list
            Array of Player objects matching the query.
        """

        VerifyPlatform(platform)

        data: dict = (await self.http.SearchPlayer(platform.value, username))["data"]

        limit: int = kwargs.get("limit", 0)
        if limit > 0:
            data = data[:limit]

        results: List[Player] = []

        for player in data:
            # The API returns the accountId as a string
            accountId: int = int(player.get("accountId"))

            avatar: Union[dict, str] = player.get("avatar")
            if isinstance(avatar, dict):
                avatar: Union[dict, str] = avatar["avatarUrlLargeSsl"]

            data = {
                "platform": player["platform"],
                "username": player["username"],
                "accountId": accountId,
                "avatarUrl": avatar,
            }

            results.append(Player(self, data))

        return results

    async def GetPlayerProfile(
        self, platform: Platform, username: str, title: Title, mode: Mode
    ) -> dict:
        """
        Get a Call of Duty player's profile for the specified title and mode.

        Parameters
        ----------
        platform : callofduty.Platform
            Platform to get the player from.
        username : str
            Player's username for the designated platform.
        title : callofduty.Title
            Call of Duty title to get the player's profile from.
        mode: callofduty.Mode
            Call of Duty mode to get the player's profile from.

        Returns
        -------
        dict
            JSON data of the player's complete profile for the requested
            title and mode.
        """

        VerifyPlatform(platform)
        VerifyTitle(title)
        VerifyMode(mode, title)

        return (
            await self.http.GetPlayerProfile(
                platform.value, username, title.value, mode.value
            )
        )["data"]

    async def GetMatch(self, title: Title, platform: Platform, matchId: int) -> Match:
        """
        Get a Call of Duty match using its title, platform, mode, and ID.

        Parameters
        ----------
        title : callofduty.Title
            Call of Duty title which the match occured on.
        platform : callofduty.Platform
            Platform to get the player from.
        matchId : int
            Match ID.

        Returns
        -------
        object
            Match object representing the specified details.
        """

        VerifyTitle(title)
        VerifyPlatform(platform)

        return Match(
            self, {"id": matchId, "platform": platform.value, "title": title.value,},
        )

    async def GetPlayerMatches(
        self, platform: Platform, username: str, title: Title, mode: Mode, **kwargs
    ) -> List[Match]:
        """
        Get a Call of Duty player's match history for the specified title and mode.

        Parameters
        ----------
        platform : callofduty.Platform
            Platform to get the player from.
        username : str
            Player's username for the designated platform.
        title : callofduty.Title
            Call of Duty title to get the player's matches from.
        mode: callofduty.Mode
            Call of Duty mode to get the player's matches from.
        limit : int, optional
            Number of matches which will be returned (default is 10.)
        startTimestamp : int, optional
            Unix timestamp representing the earliest time which a returned
            match should've occured (default is None.)
        endTimestamp : int, optional
            Unix timestamp representing the latest time which a returned
            match should've occured (default is None.)

        Returns
        -------
        list
            Array of Match objects.
        """

        VerifyPlatform(platform)
        VerifyTitle(title)
        VerifyMode(mode, title)

        limit: int = kwargs.get("limit", 10)
        startTimestamp: int = kwargs.get("startTimestamp", 0)
        endTimestamp: int = kwargs.get("endTimestamp", 0)

        if platform == Platform.Activision:
            # The preferred matches endpoint does not currently support
            # the Activision (uno) platform.
            data: dict = (
                await self.http.GetPlayerMatchesDetailed(
                    platform.value,
                    username,
                    title.value,
                    mode.value,
                    limit,
                    startTimestamp,
                    endTimestamp,
                )
            )["data"]["matches"]

            matches: List[Match] = []

            for _match in data:
                matches.append(
                    Match(
                        self,
                        {
                            # The API returns the matchId as a string
                            "id": int(_match["matchID"]),
                            "platform": platform.value,
                            "title": title.value,
                        },
                    )
                )
        else:
            data: dict = (
                await self.http.GetPlayerMatches(
                    platform.value,
                    username,
                    title.value,
                    mode.value,
                    limit,
                    startTimestamp,
                    endTimestamp,
                )
            )["data"]

            matches: List[Match] = []

            for _match in data:
                matches.append(
                    Match(
                        self,
                        {
                            # The API returns the matchId as a string
                            "id": int(_match["matchId"]),
                            "platform": platform.value,
                            "title": title.value,
                        },
                    )
                )

        return matches

    async def GetPlayerMatchesSummary(
        self, platform: Platform, username: str, title: Title, mode: Mode, **kwargs
    ) -> dict:
        """
        Get a Call of Duty player's match history summary for the specified title and mode.

        Parameters
        ----------
        platform : callofduty.Platform
            Platform to get the player from.
        username : str
            Player's username for the designated platform.
        title : callofduty.Title
            Call of Duty title to get the player's matches from.
        mode: callofduty.Mode
            Call of Duty mode to get the player's matches from.
        limit : int, optional
            Number of matches which will be returned (default is 10.)
        startTimestamp : int, optional
            Unix timestamp representing the earliest time which a returned
            match should've occured (default is None.)
        endTimestamp : int, optional
            Unix timestamp representing the latest time which a returned
            match should've occured (default is None.)

        Returns
        -------
        dict
            JSON data containing recent matches summary.
        """

        VerifyPlatform(platform)
        VerifyTitle(title)
        VerifyMode(mode, title)

        limit: int = kwargs.get("limit", 10)
        startTimestamp: int = kwargs.get("startTimestamp", 0)
        endTimestamp: int = kwargs.get("endTimestamp", 0)

        return (
            await self.http.GetPlayerMatchesDetailed(
                platform.value,
                username,
                title.value,
                mode.value,
                limit,
                startTimestamp,
                endTimestamp,
            )
        )["data"]["summary"]

    async def GetMatchDetails(
        self, title: Title, platform: Platform, matchId: int
    ) -> dict:
        """
        Get a Call of Duty match's details.

        Parameters
        ----------
        title : callofduty.Title
            Call of Duty title which the match occured on.
        platform : callofduty.Platform
            Platform to get the match from.
        matchId : int
            Match ID.

        Returns
        -------
        dict
            JSON data containing the full details of the requested Call of Duty match.
        """

        VerifyPlatform(platform)
        VerifyTitle(title)

        return (await self.http.GetMatch(title.value, platform.value, matchId))["data"]

    async def GetMatchTeams(
        self, title: Title, platform: Platform, matchId: int
    ) -> list:
        """
        Get the teams which played in a Call of Duty match.

        Parameters
        ----------
        title : callofduty.Title
            Call of Duty title which the match occured on.
        platform : callofduty.Platform
            Platform to get the match from.
        matchId : int
            Match ID.

        Returns
        -------
        list
            Array containing two child arrays, one for each team. Each
            team's array contains Player objects which represent the
            players on the team.
        """

        VerifyPlatform(platform)
        VerifyTitle(title)

        data: dict = (await self.http.GetMatch(title.value, platform.value, matchId))[
            "data"
        ]["teams"]

        # The API does not state which team is allies/axis, so no array
        # keys will be used.
        teams: list = []

        for team in data:
            # Current team iterator
            i: List[Player] = []

            for player in team:
                i.append(
                    Player(
                        self,
                        {
                            "platform": player["provider"],
                            "username": player["username"],
                            "accountId": player["unoId"],
                        },
                    )
                )

            teams.append(i)

        return teams

    async def GetLeaderboard(
        self, title: Title, platform: Platform, **kwargs
    ) -> Leaderboard:
        """
        Get a Call of Duty leaderboard.

        Parameters
        ----------
        title : callofduty.Title
            Call of Duty title which the leaderboard represents.
        platform : callofduty.Platform
            Platform to get which the leaderboard represents.
        gameType : callofduty.GameType, optional
            Game type to get the leaderboard for (default is Core.)
        gameMode : str, optional
            Game mode to get the leaderboard for (default is Career.)
        timeFrame : callofduty.TimeFrame, optional
            Time Frame to get the leaderboard for (default is All-Time.)
        page : int, optional
            Leaderboard page to get (default is 1.)

        Returns
        -------
        object
            Leaderboard object representing the specified details.
        """

        gameType: GameType = kwargs.get("gameType", GameType.Core)
        gameMode: str = kwargs.get("gameMode", "career")
        timeFrame: TimeFrame = kwargs.get("timeFrame", TimeFrame.AllTime)
        page: int = kwargs.get("page", 1)

        VerifyTitle(title)
        VerifyPlatform(platform)
        VerifyGameType(gameType)
        VerifyTimeFrame(timeFrame)

        data: dict = (
            await self.http.GetLeaderboard(
                title.value,
                platform.value,
                gameType.value,
                gameMode,
                timeFrame.value,
                page,
            )
        )["data"]

        # Leaderboard responses don't include the timeFrame, so we'll
        # just add it manually.
        data["timeFrame"] = timeFrame.value

        return Leaderboard(self, data)

    async def GetPlayerLeaderboard(
        self, title: Title, platform: Platform, username: str, **kwargs
    ) -> Leaderboard:
        """
        Get a Call of Duty leaderboard.

        Parameters
        ----------
        title : callofduty.Title
            Call of Duty title which the leaderboard represents.
        platform : callofduty.Platform
            Platform to get which the leaderboard represents.
        username : str
            Player's username for the designated platform.
        gameType : callofduty.GameType, optional
            Game type to get the leaderboard for (default is Core.)
        gameMode : str, optional
            Game mode to get the leaderboard for (default is Career.)
        timeFrame : callofduty.TimeFrame, optional
            Time Frame to get the leaderboard for (default is All-Time.)

        Returns
        -------
        object
            Leaderboard object representing the specified details.
        """

        gameType: GameType = kwargs.get("gameType", GameType.Core)
        gameMode: str = kwargs.get("gameMode", "career")
        timeFrame: TimeFrame = kwargs.get("timeFrame", TimeFrame.AllTime)

        VerifyTitle(title)
        VerifyPlatform(platform)
        VerifyGameType(gameType)
        VerifyTimeFrame(timeFrame)

        data: dict = (
            await self.http.GetPlayerLeaderboard(
                title.value,
                platform.value,
                username,
                gameType.value,
                gameMode,
                timeFrame.value,
            )
        )["data"]

        # Leaderboard responses don't include the timeFrame, so we'll
        # just add it manually.
        data["timeFrame"] = timeFrame.value

        return Leaderboard(self, data)

    async def GetLeaderboardPlayers(
        self, title: Title, platform: Platform, **kwargs
    ) -> List[Player]:
        """
        Get the players from a Call of Duty leaderboard.

        Parameters
        ----------
        title : callofduty.Title
            Call of Duty title which the leaderboard represents.
        platform : callofduty.Platform
            Platform to get which the leaderboard represents.
        gameType : callofduty.GameType, optional
            Game type to get the leaderboard for (default is Core.)
        gameMode : str, optional
            Game mode to get the leaderboard for (default is Career.)
        timeFrame : callofduty.TimeFrame, optional
            Time Frame to get the leaderboard for (default is All-Time.)
        page : int, optional
            Leaderboard page to get (default is 1.)

        Returns
        -------
        list
            Array containing Player objects for each Leaderboard entry.
        """

        gameType: GameType = kwargs.get("gameType", GameType.Core)
        gameMode: str = kwargs.get("gameMode", "career")
        timeFrame: TimeFrame = kwargs.get("timeFrame", TimeFrame.AllTime)
        page: int = kwargs.get("page", 1)

        VerifyTitle(title)
        VerifyPlatform(platform)
        VerifyGameType(gameType)
        VerifyTimeFrame(timeFrame)

        data: dict = (
            await self.http.GetLeaderboard(
                title.value,
                platform.value,
                gameType.value,
                gameMode,
                timeFrame.value,
                page,
            )
        )["data"]

        # Leaderboard responses don't include the timeFrame, so we'll
        # just add it manually.
        data["timeFrame"] = timeFrame.value

        players: List[Player] = []
        for entry in Leaderboard(self, data).entries:
            players.append(
                Player(self, {"platform": platform.value, "username": entry.username})
            )

        return players

    async def GetAvailableMaps(
        self,
        title: Title,
        platform: Platform = Platform.PlayStation,
        mode: Mode = Mode.Multiplayer,
    ) -> list:
        """
        Get the Maps available in the specified Title for Heat Map use.

        Parameters
        ----------
        title : callofduty.Title
            Call of Duty title which the maps originate.
        platform : callofduty.Platform, optional
            Platform which the maps are available on (default is PlayStation.)
        mode: callofduty.Mode, optional
            Call of Duty mode to get the maps from (default is Multiplayer.)

        Returns
        -------
        list
            Array of Maps and the Game Modes which are available each map.
        """

        return (
            await self.http.GetAvailableMaps(title.value, platform.value, mode.value)
        )["data"]

    async def GetLootSeason(self, title: Title, season: int, **kwargs) -> Season:
        """
        Get a Call of Duty Loot Season by its title and number.

        Parameters
        ----------
        title : callofduty.Title
            Call of Duty title which the loot season originates.
        season : int
            Loot season number relative to title.
        platform : callofduty.Platform, optional
            Platform which the loot season is available on (default is PlayStation.)
        language : callofduty.Language, optional
            Language which the loot data should be in (default is English.)

        Returns
        -------
        object
            Season object for the requested Loot Season.
        """

        platform = kwargs.get("platform", Platform.PlayStation)
        language = kwargs.get("language", Language.English)

        VerifyPlatform(platform)
        VerifyLanguage(language)

        data = (
            await self.http.GetLootSeason(
                title.value, season, platform.value, language.value
            )
        )["data"]

        # Loot Season responses don't include these values, so we'll just
        # add them manually.
        data["title"] = title.value
        data["platform"] = platform.value
        data["season"] = season
        data["language"] = language.value

        return Season(self, data)

    async def GetSquad(self, name: str) -> Squad:
        """
        Get a Call of Duty Squad using its name.

        Parameters
        ----------
        name : str
            Name of Squad.

        Returns
        -------
        object
            Squad object for the requested Squad.
        """

        return Squad(self, (await self.http.GetSquad(name))["data"])

    async def GetPlayerSquad(self, platform: Platform, username: str) -> Squad:
        """
        Get a Call of Duty player's Squad using their platform and username.

        Parameters
        ----------
        platform : callofduty.Platform
            Platform to get the player from.
        username : str
            Player's username for the designated platform.

        Returns
        -------
        object
            Squad object for the requested Squad.
        """

        VerifyPlatform(platform)

        return Squad(
            self, (await self.http.GetPlayerSquad(platform.value, username))["data"]
        )

    async def GetMySquad(self) -> Squad:
        """
        Get the Squad of the authenticated Call of Duty player.

        Returns
        -------
        object
            Squad object for the requested Squad.
        """

        return Squad(self, (await self.http.GetMySquad())["data"])

    async def JoinSquad(self, name: str):
        """
        Join a Call of Duty Squad using its name.

        Parameters
        ----------
        name : str
            Name of Squad.
        """

        await self.http.JoinSquad(name)

    async def LeaveSquad(self) -> Squad:
        """
        Leave the Call of Duty Squad of the authenticated player.
        Upon leaving a Squad, the player is automatically placed into
        a random Squad.

        Returns
        -------
        object
            Squad object for the randomly-joined Squad.
        """

        await self.http.LeaveSquad()

        return await self.GetMySquad()



"""
CallofDuty.py is an asynchronous, object-oriented Python wrapper for
the Call of Duty API.

GitHub: https://github.com/EthanC/CallofDuty.py
"""

import logging

from .auth import Login
from .client import Client
from .enums import *
from .errors import *
from .leaderboard import Leaderboard
from .loot import LootItem, Season
from .match import Match
from .player import Player
from .squad import Squad

try:
    from logging import NullHandler
except ImportError:

    class NullHandler(logging.Handler):
        def emit(self, record):
            pass


logging.getLogger(__name__).addHandler(NullHandler())



import logging
import re

from .enums import GameType, Language, Mode, Platform, TimeFrame, Title
from .errors import (
    InvalidGameType,
    InvalidLanguage,
    InvalidMode,
    InvalidPlatform,
    InvalidTimeFrame,
    InvalidTitle,
)

log: logging.Logger = logging.getLogger(__name__)


def VerifyPlatform(value: Platform):
    """
    Raise an InvalidPlatform client exception if a value which is not
    present in the Platform enum is passed.

    Parameters
    ----------
    value : callofduty.Platform
        Value to confirm is present in the Platform enum.
    """

    if value not in Platform:
        raise InvalidPlatform(f"{value.name} is not a valid platform")


def VerifyTitle(value: Title):
    """
    Raise an InvalidTitle client exception if a value which is not
    present in the Title enum is passed.

    Parameters
    ----------
    value : callofduty.Title
        Value to confirm is present in the Title enum.
    """

    if value not in Title:
        raise InvalidTitle(f"{value.name} is not a valid title")


def VerifyMode(value: Mode, title: Title):
    """
    Raise an InvalidMode client exception if a value which is not
    present in the Mode enum is passed.

    Parameters
    ----------
    value : callofduty.Mode
        Value to confirm is present in the Mode enum.
    title : callofduty.Title
        Title to confirm is compatible with the Mode.
    """

    if value not in Mode:
        raise InvalidMode(f"{value.name} is not a valid mode")
    elif (value == Mode.Zombies) and (title == Title.ModernWarfare):
        raise InvalidMode(f"{value.name} is not a valid mode for title {title.name}")
    elif (value == Mode.Blackout) and (title != Title.BlackOps4):
        raise InvalidMode(f"{value.name} is not a valid mode for title {title.name}")


def VerifyLanguage(value: Language):
    """
    Raise an InvalidLanguage client exception if a value which is not
    present in the Language enum is passed.

    Parameters
    ----------
    value : callofduty.Language
        Value to confirm is present in the Language enum.
    """

    if value not in Language:
        raise InvalidLanguage(f"{value.name} is not a valid language")


def VerifyTimeFrame(value: TimeFrame):
    """
    Raise an InvalidTimeFrame client exception if a value which is not
    present in the TimeFrame enum is passed.

    Parameters
    ----------
    value : callofduty.TimeFrame
        Value to confirm is present in the TimeFrame enum.
    """

    if value not in TimeFrame:
        raise InvalidTimeFrame(f"{value.name} is not a valid time frame")


def VerifyGameType(value: GameType):
    """
    Raise an InvalidGameType client exception if a value which is not
    present in the GameType enum is passed.

    Parameters
    ----------
    value : callofduty.GameType
        Value to confirm is present in the GameType enum.
    """

    if value not in GameType:
        raise InvalidGameType(f"{value.name} is not a valid game type")


def StripHTML(input: str) -> str:
    """
    Strip the HTML formatting from a string.

    Parameters
    ----------
    input : str
        HTML formatted string.

    Returns
    -------
    str
        Input string without the HTML formatting.
    """

    # Regex is hideous, but this gets the job done faster than an external
    # library. This is also future-proof against any sort of HTML characters,
    # such as &nbsp and &amp.
    expression: re.Pattern[str] = re.compile(
        "<.*?>|&([a-z0-9]+|#[0-9]{1,6}|#x[0-9a-f]{1,6});"
    )

    return re.sub(expression, "", input)



import logging
import random
from typing import Dict, Optional

import httpx

from .client import Client
from .errors import LoginFailure
from .http import HTTP

log: logging.Logger = logging.getLogger(__name__)


class Auth:
    """
    Implements the Call of Duty authorization flow.

    Parameters
    ----------
    email : str
        Activision account email address.
    password : str
        Activision account password.
    """

    loginUrl: str = "https://profile.callofduty.com/cod/mapp/login"
    registerDeviceUrl: str = "https://profile.callofduty.com/cod/mapp/registerDevice"

    _accessToken: Optional[str] = None
    _deviceId: Optional[str] = None

    def __init__(self, email: str, password: str):
        self.email: str = email
        self.password: str = password

        self.session: httpx.AsyncClient = httpx.AsyncClient()

    @property
    def AccessToken(self) -> Optional[str]:
        """
        Returns
        -------
        str
            Access Token which is set during the device registration phase
            of authentication.
        """

        if self._accessToken is None:
            raise LoginFailure("Access Token is null, not authenticated")

        return self._accessToken

    @property
    def DeviceId(self) -> Optional[str]:
        """
        Returns
        -------
        str
            Device ID which is set during the device registration phase
            of authentication.
        """

        if self._deviceId is None:
            raise LoginFailure("DeviceId is null, not authenticated")

        return self._deviceId

    async def RegisterDevice(self):
        """
        Generate and register a Device ID with the Call of Duty API. Set
        the corresponding Access Token if successful.
        """

        self._deviceId: Optional[str] = hex(random.getrandbits(128)).lstrip("0x")

        body: Dict[str, Optional[str]] = {"deviceId": self.DeviceId}

        async with self.session as client:
            res: httpx.Response = await client.post(self.registerDeviceUrl, json=body)

            if res.status_code != 200:
                raise LoginFailure(
                    f"Failed to register fake device (HTTP {res.status_code})"
                )

            data: dict = res.json()

            self._accessToken: Optional[str] = data["data"]["authHeader"]

    async def SubmitLogin(self):
        """
        Submit the specified login credentials to the Call of Duty API using the
        previously acquired Access Token and Device ID.
        """

        headers: Dict[str, str] = {
            "Authorization": f"Bearer {self.AccessToken}",
            "x_cod_device_id": self.DeviceId,
        }

        data: Dict[str, str] = {"email": self.email, "password": self.password}

        async with self.session as client:
            res: httpx.Response = await client.post(
                self.loginUrl, json=data, headers=headers
            )

            if res.status_code != 200:
                raise LoginFailure(f"Failed to login (HTTP {res.status_code})")


async def Login(email: str, password: str) -> Client:
    """
    Convenience function to make login with the Call of Duty authorization flow
    as easy as possible.

    Parameters
    ----------
    email : str
        Activision account email address.
    password : str
        Activision account password.

    Returns
    -------
    object
        Authenticated Call of Duty client.
    """

    auth: Auth = Auth(email, password)

    await auth.RegisterDevice()
    await auth.SubmitLogin()

    return Client(HTTP(auth))



from typing import Union


class CallofDutyException(Exception):
    """Base exception class for CallofDuty.py"""

    pass


class ClientException(CallofDutyException):
    """Exception which is thrown when an operation in the Client class fails."""

    pass


class LoginFailure(ClientException):
    """
    Exception which is thrown when authentication fails due to incorrect
    credentials or an unknown error.
    """

    pass


class InvalidPlatform(ClientException):
    """
    Exception which is thrown when an invalid platform is passed to any
    client function.
    """

    pass


class InvalidTitle(ClientException):
    """
    Exception which is thrown when an invalid title is passed to any
    client function.
    """

    pass


class InvalidMode(ClientException):
    """
    Exception which is thrown when an invalid mode is passed to any
    client function.
    """

    pass


class InvalidLanguage(ClientException):
    """
    Exception which is thrown when an invalid language is passed to any
    client function.
    """

    pass


class InvalidTimeFrame(ClientException):
    """
    Exception which is thrown when an invalid time frame is passed to any
    client function.
    """

    pass


class InvalidGameType(ClientException):
    """
    Exception which is thrown when an invalid game type is passed to any
    client function.
    """

    pass


class HTTPException(CallofDutyException):
    """
    Exception which is thrown when an HTTP request operation fails.

    Parameters
    ----------
    statusCode : int
        HTTP status code of the request.
    res : dict
        Response of the HTTP request.
    """

    def __init__(self, statusCode: int, res: Union[dict, str]):
        if isinstance(res, dict):
            try:
                message: Union[dict, str] = res["data"].get("message", res)
            except AttributeError:
                # This allows us to capture reasoning from the Squads
                # endpoints which don't follow the usual response structure.
                message: Union[dict, str] = res.get("data", res)
        else:
            message: Union[dict, str] = res

        super().__init__(f"HTTP {statusCode} - {message}")


class Forbidden(HTTPException):
    """Exception which is thrown when HTTP status code 403 occurs."""

    pass


class NotFound(HTTPException):
    """Exception which is thrown when HTTP status code 404 occurs."""

    pass



import logging
from typing import Optional

from .client import Client

log: logging.Logger = logging.getLogger(__name__)


class Object:
    """
    Represents a generic Call of Duty object.

    Parameters
    ----------
    client : callofduty.Client
        Client which manages communication with the Call of Duty API.
    """

    _type: Optional[str] = None

    def __init__(self, client: Client):
        self._client: Client = client

    @property
    def type(self) -> Optional[str]:
        return self._type

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"

    def __str__(self) -> str:
        return self.__repr__()



import logging
from typing import Dict, List, Union

from .enums import GameType, Platform, TimeFrame, Title
from .object import Object

log: logging.Logger = logging.getLogger(__name__)


class Leaderboard(Object):
    """
    Represents a Call of Duty leaderboard object.

    Parameters
    ----------
    title : callofduty.Title
        Call of Duty title which the leaderboard represents.
    platform : callofduty.Platform
        Platform to get which the leaderboard represents.
    gameType : callofduty.GameType, optional
        Game type to get the leaderboard for (default is Core.)
    gameMode : str, optional
        Game mode to get the leaderboard for (default is Career.)
    timeFrame : callofduty.TimeFrame, optional
        Time Frame to get the leaderboard for (default is All-Time.)
    page : int, optional
        Leaderboard page to get (default is 1.)
    pages : int, optional
        Total number of pages available for the leaderboard.
    columns : list, optional
        Array of strings containing the column headers for the leaderboard.
    entries : list, optional
        Array of Leaderboard Entry objects containing the entries for the leaderboard.
    """

    _type: str = "Leaderboard"

    def __init__(self, client, data: dict):
        super().__init__(client)

        self.title: Title = Title(data.pop("title"))
        self.platform: Platform = Platform(data.pop("platform"))
        self.gameType: GameType = GameType(data.pop("leaderboardType"))
        self.gameMode: str = data.pop("gameMode")
        self.timeFrame: TimeFrame = TimeFrame(data.pop("timeFrame"))
        self.page: int = data.pop("page")
        self.pages: int = data.pop("totalPages")
        self.columns: list = data.pop("columns")
        self.entries: List[LeaderboardEntry] = []

        for entry in data.pop("entries", []):
            # Leaderboard Entries don't include this value, so we'll just
            # add it manually.
            entry["platform"] = self.platform.value

            self.entries.append(LeaderboardEntry(self, entry))

    async def players(self) -> list:
        """
        Get the players from a Call of Duty leaderboard.

        Returns
        -------
        list
            Array of Player objects for each leaderboard entry.
        """

        return await self._client.GetLeaderboardPlayers(
            self.title,
            self.platform,
            gameType=self.gameType,
            gameMode=self.gameMode,
            timeFrame=self.timeFrame,
            page=self.page,
        )


class LeaderboardEntry(Object):
    """
    Represents a Call of Duty leaderboard entry object.

    Parameters
    ----------
    platform : callofduty.Platform
        Platform of the player.
    username : str
        Player's username for the designated platform.
    rank : int
        Leaderboard position for the designated entry.
    updated : int, optional
        Value in seconds representing how long ago the leaderboard entry was updated.
    rating : int, optional
        Unknown rating value.
    values : dict, optional
        JSON data containing values for the leaderboard entry.
    """

    _type: str = "LeaderboardEntry"

    def __init__(self, client, data: dict):
        super().__init__(client)

        self.platform: Platform = Platform(data.pop("platform"))
        self.username: str = data.pop("username")
        self.rank: int = int(data.pop("rank"))
        self.updated: int = int(data.pop("updateTime"))
        self.rating: int = data.pop("rating")
        self.values: Dict[str, Union[int, float]] = data.pop("values")
